import re
import json
from pathlib import Path
from copy import deepcopy

import yaml
from yaml.scanner import MarkedYAMLError

from .common import HarrierProblem, logger

DEFAULT_CONFIG = Path(__file__).parent / 'harrier.default.yml'


# in order if preference:
DEFAULT_CONFIG_FILES = [
    'harrier.yml',
    'harrier.json',
    'config.yml',
    'config.json',
]


class Config:
    _already_setup = _base_dir = _target = target_dir = served_direct = None

    def __init__(self, file_path=None, config_dict: dict=None):
        if config_dict is None:
            config_dict, file_path = self._load_file(file_path)
        self.config_file = file_path or Path('./none)')
        self._orig_config = config_dict
        self._config = self._prepare_config(config_dict)
        self.root = Path(self._config['root'])

    @classmethod
    def _load_file(cls, config_file):
        if config_file:
            config_file = Path(config_file)
            if config_file.is_file():
                file_path = config_file
            else:
                file_path = cls._find_config_file(config_file)
        else:
            file_path = cls._find_config_file()

        # TODO we should probably keep config_file as None
        config, config_file = {}, None
        if file_path is None:
            msg = 'no config file supplied and none found with expected names: {}. Using default settings.'
            logger.warning(msg.format(', '.join(DEFAULT_CONFIG_FILES)))
        else:
            loader, _ = yaml_or_json(str(file_path))
            with file_path.open() as f:
                try:
                    config = loader(f)
                except (MarkedYAMLError, ValueError) as e:
                    logger.error('%s: %s', e.__class__.__name__, e)
                    raise HarrierProblem('error loading "{}"'.format(file_path)) from e
            config_file = file_path.relative_to('.')
        return config, config_file

    @classmethod
    def _find_config_file(cls, path=Path('.')):
        logger.debug('looking for config file with default name in "%s"', path.relative_to('.'))
        for default_file in DEFAULT_CONFIG_FILES:
            for fn in path.iterdir():
                if fn.name == default_file:
                    logger.info('Found default config file {}'.format(fn.name))
                    return fn

    def _prepare_config(self, config):
        with DEFAULT_CONFIG.open() as f:
            c = yaml.load(f)
        unknown = set(config.keys()) - set(c.keys())
        if unknown:
            raise HarrierProblem('Unexpected sections in config: {}'.format(unknown))
        return self._merge_dicts(config, c)

    @classmethod
    def _merge_dicts(cls, d_update:  dict, d_base:  dict):
        d_new = deepcopy(d_base)
        for k, v in d_update.items():
            base_v = d_base.get(k)
            if isinstance(v, dict) and isinstance(base_v, dict):
                d_new[k] = cls._merge_dicts(v, base_v)
            else:
                d_new[k] = v
        return d_new

    def setup(self, target_name='build', served_direct=False, base_dir=None):
        if self._already_setup:
            return
        self.root = self._find_root_dir(base_dir)
        logger.debug('Full root directory %s exists ✓', self.root)
        self.config_file = self.config_file.relative_to(self._base_dir)
        self._set_target(target_name)
        self._already_setup = True
        self.served_direct = served_direct

    def _set_target(self, name):
        target = self._config['target']
        self._target = target.get(name) or {}
        self._target['name'] = name

        default_paths = {
            'build': 'build',
            'serve': '.serve',
        }

        target_dir = self._target.get('path') or default_paths.get(name, default_paths['serve'])
        target_dir = self._base_dir / target_dir
        if not target_dir.parent.exists():
            raise HarrierProblem('parent of target directory {} does not exist'.format(target_dir))
        self.target_dir = target_dir / self.subdirectory
        logger.debug('Output directory set to %s ✓', self.target_dir)

    def _find_root_dir(self, base_dir):
        self._base_dir = Path(base_dir or self.config_file.parent)
        logger.debug('Setting config root directory relative to {}'.format(self._base_dir))
        try:
            root_dir = self._base_dir.joinpath(self.root).resolve()
        except FileNotFoundError:
            if base_dir is None:
                msg = 'config root "{root}" does not exist relative to config file directory "{base_dir}"'
            else:
                msg = 'config root "{root}" does not exist relative to directory "{base_dir}"'
            raise HarrierProblem(msg.format(root=self.root, base_dir=self._base_dir))
        else:
            return root_dir

    def _get_setting(self, *args):

        def find_property(d, key_list):
            # TODO show more helpful errors eg if v is a string and v.get is called.
            v = deepcopy(d)
            for k in key_list:
                v = v.get(k)
                if v is None:
                    return
            return v

        target_prop = find_property(self._target, args)
        if target_prop is not None:
            return target_prop
        return find_property(self._config, args)

    @property
    def subdirectory(self):
        sd = self._get_setting('subdirectory')
        if sd.startswith('/'):
            raise HarrierProblem('subdirectory should be relative and not start with "/"')
        return Path(sd)

    @property
    def uri_subdirectory(self):
        s = str(self.subdirectory).strip('/')
        return '/{}/'.format(s) if s else '/'

    @property
    def serve_port(self):
        return self._target.get('port') or 8000

    @property
    def serve_livereload(self):
        return self._target.get('livereload', True)

    @property
    def jinja_directories(self):
        rel_dirs = self._listify(self._get_setting('jinja', 'directories'))
        dirs = []
        for rel_dir in rel_dirs:
            full_dir = self.root / rel_dir
            if not full_dir.exists():
                raise HarrierProblem('"{}" does not exist'.format(full_dir))
            elif not full_dir.is_dir():
                raise HarrierProblem('"{}" is not a directory'.format(full_dir))
            dirs.append(str(full_dir))
        return dirs

    @property
    def jinja_patterns(self):
        return self._get_setting('jinja', 'patterns') + self._get_setting('jinja', 'extra_patterns')

    @property
    def sass_precision(self):
        return int(self._get_setting('sass', 'precision'))

    @property
    def execute_commands(self):
        return self._listify(self._get_setting('execute', 'commands'))

    @property
    def execute_patterns(self):
        return self._listify(self._get_setting('execute', 'patterns'))

    @property
    def execute_cleanup(self):
        cleanup = self._get_setting('execute', 'cleanup')
        return True if cleanup is None else bool(cleanup)

    @property
    def asset_file(self):
        return self._get_setting('assets', 'file')

    @property
    def asset_url_root(self):
        url_root = self._get_setting('assets', 'url_root')
        if url_root is None:
            return self.uri_subdirectory
        if re.match('^https?://', url_root) and not re.match('^https?://.*?/.+', url_root):
            # url_root does not include path so subdirectory should be appended
            return url_root.rstrip('/') + self.uri_subdirectory
        return url_root

    @property
    def path_mapping(self):
        # TODO deal better with conf_dict, eg. dicts, list of dicts, check length of lists of lists
        # TODO add extra_mapping to avoid overriding these values
        return self._config['mapping']

    @property
    def exclude_patterns(self):
        exclude = set(self._config['exclude'])
        return exclude | set('*/' + e for e in exclude)

    @property
    def tools(self):
        tools = []
        for k, v in self._config.items():
            if isinstance(v, dict) and 'tool_path' in v and v.get('active', False) is True:
                tools.append(v['tool_path'])
        return tools

    @property
    def context(self):
        _ctx = {
            'build_target': self._target['name'],
        }
        _ctx.update(self._config['context'])
        _ctx.update(self._target.get('context') or {})
        return _ctx

    @property
    def subprocesses(self):
        return self._config['subprocesses']

    def find_library(self):
        ldir = self._config['library']
        dirs = [
            self.root / ldir,
            self._base_dir / ldir,
            Path(ldir),
        ]
        for d in dirs:
            if d.exists():
                logger.debug('Found library directory {}'.format(d))
                return d

        if 'library' in self._orig_config:
            raise HarrierProblem("library supplied in config but can't be found.")
        else:
            logger.debug('unable to find library directory, ignoring as library was not supplied in config')

    def _listify(self, v):
        if isinstance(v, str):
            return [v]
        return list(v or [])


def pretty_yaml_dump(obj):
    return yaml.dump(obj, default_flow_style=False)


def pretty_json_dump(obj):
    return json.dumps(obj, indent=2, sort_keys=True) + '\n'


def yaml_or_json(file_path):
    if any(file_path.endswith(ext) for ext in ['.yaml', '.yml']):
        logger.debug('Processing %s as a yaml file', file_path)
        return yaml.load, pretty_yaml_dump
    elif file_path.endswith('.json'):
        logger.debug('Processing %s as a json file', file_path)
        return json.load, pretty_json_dump
    else:
        msg = 'Unexpected extension for "{}", should be json or yml/yaml'
        raise HarrierProblem(msg.format(file_path))



import click
import re
import logging

from harrier import VERSION
from .build import build
from .config import Config
from .common import logger, HarrierProblem
from .watch import watch

config_help = 'Provide a specific harrier config yml file path.'
dev_address_help = 'IP address and port to serve documentation locally (default: localhost:8000).'
target_help = 'choice from targets in harrier.yml, defaults to same value as action eg. build or serve.'
site_dir_help = 'The directory to output the result of the documentation build.'
reload_help = 'Enable and disable the live reloading in the development server.'
verbose_help = 'Enable verbose output.'


class ClickHandler(logging.Handler):
    colours = {
        logging.DEBUG: 'white',
        logging.INFO: 'green',
        logging.WARN: 'yellow',
    }

    def emit(self, record):
        log_entry = self.format(record)
        colour = self.colours.get(record.levelno, 'red')
        m = re.match('^(\[.*?\])', log_entry)
        if m:
            time = click.style(m.groups()[0], fg='magenta')
            msg = click.style(log_entry[m.end():], fg=colour)
            click.echo(time + msg)
        else:
            click.secho(log_entry, fg=colour)


def setup_logging(verbose=False, times=False):
    for h in logger.handlers:
        if isinstance(h, ClickHandler):
            return
    handler = ClickHandler()
    fmt = '[%(asctime)s] %(message)s' if times else '%(message)s'
    formatter = logging.Formatter(fmt, datefmt='%H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)


@click.command()
@click.version_option(VERSION, '-V', '--version')
@click.argument('action', type=click.Choice(['serve', 'build']))
@click.argument('config-file', type=click.Path(exists=True), required=False)
@click.option('-t', '--target', help=dev_address_help)
@click.option('-a', '--dev-addr', help=dev_address_help, metavar='<IP:PORT>')
@click.option('-v', '--verbose', is_flag=True, help=verbose_help)
def cli(action, config_file, target, dev_addr, verbose):
    """
    harrier - Jinja2 & sass/scss aware site builder
    """
    is_live = action == 'serve'  # TODO add watch
    is_served = action == 'serve'
    setup_logging(verbose, times=is_live)
    try:
        config = Config(config_file)
        target = target or action
        config.setup(target, served_direct=is_served)
        if action == 'serve':
            watch(config)
        else:
            assert action == 'build'
            build(config)
    except HarrierProblem as e:
        msg = 'Error: {}'
        if not verbose:
            msg += ', use "--verbose" for more details'
        click.secho(msg.format(e), fg='red', err=True)



import re
import shlex
import subprocess
from copy import deepcopy
from fnmatch import fnmatch
from itertools import chain
from pathlib import PurePosixPath, Path

import yaml
import sass
from jinja2 import Environment, FileSystemLoader, contextfilter

from .config import Config, yaml_or_json
from .common import logger, HarrierProblem


def walk(root: Path):
    def gen(p):
        for _p in p.iterdir():
            if _p.is_dir():
                yield from gen(_p)
            else:
                yield _p.relative_to(root)
    return list(gen(root))


def hash_file(path: Path):
    with path.open('rb') as f:
        return hash(f.read())

FRONT_MATTER_REGEX = re.compile(r'^---[ \t]*(.*)\n---[ \t]*\n', re.S)


def parse_front_matter(s):
    m = re.match(FRONT_MATTER_REGEX, s)
    if not m:
        return None, s
    data = yaml.load(m.groups()[0]) or {}
    return data, s[m.end():]


def underscore_prefix(file_path: Path):
    return file_path.name.startswith('_')


class Tool:
    ownership_regex = None
    # priorities are used to sort tools before ownership check and build, highest comes first.
    ownership_priority = 0
    build_priority = 0
    extra_files = []
    # whether or not one file changing requires a complete rebuild
    change_sensitive = True
    single_call = False

    def __init__(self, config: Config, partial_build: bool):
        self._config = config
        self._partial = partial_build
        self.to_build = []
        self.active = False

    def assign_file(self, file_path: Path, changed: bool) -> bool:
        owned = self._check_ownership(file_path)
        if owned and (not self._partial or self.change_sensitive or changed):
            self.active |= changed
            self.to_build.append(file_path)
        return owned

    def _check_ownership(self, file_path):
        """
        Return true if this converter takes care of the given file_name, else false.

        The convert might also want to record the file to process it later.
        """
        return re.match(self.ownership_regex, str(file_path))

    def map_path(self, fp: Path):
        fps = str(fp)
        for pattern, replace in self._config.path_mapping:
            fps = re.sub(pattern, replace, fps)
        return Path(fps)

    def build(self):
        files_built = 0
        source_map = {}
        for file_path in self.to_build:
            source_files = set()
            for new_file_path, file_content in self.convert_file(file_path):
                new_file_path = new_file_path or self.map_path(file_path)
                source_files.add(new_file_path)

                target_file = Path(self._config.target_dir) / new_file_path

                if target_file.exists() and hash(file_content) == hash_file(target_file):
                    continue

                target_file.parent.mkdir(parents=True, exist_ok=True)
                with target_file.open('wb') as f:
                    f.write(file_content)
                files_built += 1
            source_map[file_path] = source_files
            if self.single_call:
                break
        return files_built, source_map

    def convert_file(self, file_path: Path) -> tuple:
        """
        generate files this converter is responsible for in the target directory.
        """
        raise NotImplementedError()

    def cleanup(self):
        pass

    @property
    def name(self):
        return self.__class__.__name__

    def __str__(self):
        return '<{} tool>'.format(self.name)


class Execute(Tool):
    ownership_priority = 10  # should go first
    build_priority = 10  # should go first
    single_call = True  # should be configurable in case command should be run for every file

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands = self._config.execute_commands
        self.ownership_patterns = self._config.execute_patterns
        self._extra_files = None

    def _check_ownership(self, file_path):
        # TODO is just str correct here?
        if not self._commands or str(file_path) in self.extra_files:
            return False
        return any(file_path.match(m) for m in self.ownership_patterns)

    def build(self):
        super().build()
        return len(self.extra_files), {}

    def convert_file(self, file_path: Path):
        for raw_command in self._commands:
            if isinstance(raw_command, dict):
                generates = raw_command.get('generates', [])
                raw_command = raw_command['command']
            else:
                generates = []
            # TODO do we need any other transforms?
            command = raw_command.format(ROOT=self._config.root)
            args = shlex.split(command)
            # TODO env variables eg. NODE_ENV
            try:
                cp = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            except FileNotFoundError as e:  # TODO any other exceptions?
                logger.error('%s: %s', e.__class__.__name__, e)
                raise HarrierProblem('problem executing "{}"'.format(command)) from e

            if cp.returncode != 0:
                logger.error('"%s" -> %s', command, cp.stdout.decode('utf8'))
                raise HarrierProblem('command "{}" returned non-zero exit status 1'.format(command))

            for path in generates:
                full_path = self._config.root / path
                if not full_path.exists():
                    logger.error('"%s" -> %s', command, cp.stdout.decode('utf8'))
                    raise HarrierProblem('command "{}" failed to generate {}'.format(raw_command, path))

            logger.debug('"%s" -> "%s" ✓', command, cp.stdout.decode('utf8'))
        return
        # noinspection PyUnreachableCode
        # we want an empty generator
        yield  # pragma: no cover

    def cleanup(self):
        if not self.active or not self._config.execute_cleanup:
            return

        for fp in self.extra_files:
            self._config.root.joinpath(fp).unlink()

    @property
    def extra_files(self):
        if self._extra_files is None:
            generates = [c.get('generates', []) for c in self._commands if isinstance(c, dict)]
            self._extra_files = list(chain(*generates))
        return self._extra_files


class Copy(Tool):
    ownership_regex = r'.*'
    ownership_priority = -10  # should go last
    change_sensitive = False

    def convert_file(self, file_path: Path):
        path = Path(self._config.root) / file_path
        try:
            path = path.resolve()
        except FileNotFoundError:
            raise HarrierProblem('"{}" does not exist'.format(path))
        else:
            with path.open('rb') as f:
                yield None, f.read()


class Sass(Tool):
    ownership_regex = r'.*\.s(a|c)ss$'
    ownership_priority = 1  # should go early

    def convert_file(self, file_path: Path):
        full_path = self._config.root / file_path
        if underscore_prefix(file_path):
            # sass files starting with underscores are partials and should not be deployed themselves
            return
        try:
            content_str = sass.compile(
                filename=str(full_path),
                precision=self._config.sass_precision,
            )
        except sass.CompileError as e:
            error = e.args[0]
            error = error.decode('utf8') if isinstance(error, bytes) else error
            error = error.strip('\n ')
            logger.error(error)
            raise HarrierProblem('Error compiling SASS') from e
        # TODO cope with maps etc.
        yield None, content_str.encode('utf8')


class FrontMatterFileSystemLoader(FileSystemLoader):
    content_cache = {}

    def get_source(self, environment, template):
        return self.content_cache[str(template)]

    def parse_source(self, environment, template):
        contents, filename, uptodate = super().get_source(environment, template)
        data, contents = parse_front_matter(contents)
        self.content_cache[template] = (contents, filename, uptodate)
        return data or {}


class Jinja(Tool):
    ownership_priority = 5  # should go early
    build_priority = -8  # should go late
    _template_files = None
    live_reload_slug = '\n<script src="http://localhost:{}/livereload.js"></script>\n'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO store the loader or env on the tool factory for faster partial builds
        # (this would need to cope with new files)
        self._loader = FrontMatterFileSystemLoader(self._config.jinja_directories)
        self._env = Environment(loader=self._loader)

        self._env.filters['S'] = self.static_file_filter

        self._file_ctxs = {}
        self._initialise_templates()

        self._ctx = self._config.context
        self._library = self._config.find_library()
        self._library_files = walk(self._library) if self._library else []
        self._extra_files = []

    def _initialise_templates(self):
        template_names = self._env.list_templates(filter_func=self._filter_template)
        self._template_files = set(template_names)
        for t in self._template_files:
            self._file_ctxs[t] = self._loader.parse_source(self._env, t)

    @contextfilter
    def static_file_filter(self, context, file_url, library=None):
        if file_url.startswith('/'):
            file_path = file_url.lstrip('/')
        else:
            file_dir = Path(context.name).parent
            file_path = file_dir / file_url

        if library:
            return self._library_file(file_path, file_url, library)
        else:
            return self._root_file(file_path, file_url)

    def _root_file(self, file_path, file_url):
        target_path = self._config.target_dir / file_path
        if target_path.exists():
            return file_url

        mapped_target_path = self.map_path(target_path)
        if mapped_target_path.exists():
            return self.map_path(file_url)
        raise HarrierProblem('unable to find {} in build directory'.format(file_path))

    def _library_file(self, file_path, file_url, library):
        lib_path = self._find_lib_file(library, file_url)
        if lib_path is None:
            raise HarrierProblem('unable to find {} in library directory, url: {}'.format(library, file_url))

        with lib_path.open('rb') as f:
            self._extra_files.append((str(file_path), f.read()))
        return file_url

    def _find_lib_file(self, file_path, file_url):
        file_path = re.sub('^libs?/', '', file_path)
        if file_path in self._library_files:
            return self._library / file_path
        for lf in self._library_files:
            lfs = str(lf)
            if lfs.endswith(file_path) or (lfs.startswith(file_path) and lfs.endswith(file_url)):
                return self._library / lf

    def _check_ownership(self, file_path):
        return str(file_path) in self._template_files

    def _filter_template(self, file_path):
        return any(fnmatch(file_path, m) for m in self._config.jinja_patterns)

    def convert_file(self, file_path: Path):
        if file_path in self._extra_files:
            return
        self._extra_files = []
        template = self._env.get_template(str(file_path))
        file_ctx = self._file_ctxs[str(file_path)]
        ctx = deepcopy(self._ctx)
        ctx.update(file_ctx)
        content_str = template.render(ctx)
        if not underscore_prefix(file_path):
            # to be consistent with sass and allow base templates to not be deployed we ignore files starting
            # with underscores
            if self._config.served_direct and self._config.serve_livereload:
                content_str += self.live_reload_slug.format(self._config.serve_port)
            yield None, content_str.encode('utf8')
        for name, content in self._extra_files:
            yield name, content


class AssetDefinition(Tool):
    ownership_priority = 0
    build_priority = -10  # should go last

    def __init__(self, config: Config, partial_build: bool):
        super().__init__(config, partial_build)
        self.to_build = [config.asset_file]
        self.active = True

    @classmethod
    def _get_commit(cls):
        try:
            cp = subprocess.run(['git', 'rev-parse', 'HEAD'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except FileNotFoundError:
            pass
        else:
            if cp.returncode == 0:
                return cp.stdout.decode('utf8').strip('\n')
        return 'unknown'

    def _check_ownership(self, file_path):
        return False

    def convert_file(self, file_path: Path):
        # TODO if this could ever be called more than once for the same target_dir we should cache the result
        _, dumper = yaml_or_json(file_path)
        commit = self._get_commit()
        file_map = {}
        root = UrlPath(self._config.asset_url_root)
        for f in walk(self._config.target_dir):
            # TODO remove file hashes from key once they're implemented
            file_map[str(f)] = str(root / f)
        # TODO, find version from previous deploy, add hash of all files this could be done with
        # modified source_map since hash doesn't have to be correct on partial builds
        obj = {
            'commit': commit,
            'files': file_map,
        }
        content = dumper(obj)
        yield file_path, content.encode('utf8')


class UrlPath(PurePosixPath):
    def __str__(self):
        s = super().__str__()
        # we have to reinstate the second / of http:// as PurePosixPath removes it
        return re.sub('(https?:)/([^/])', r'\1//\2', s)



import shutil
from copy import copy
from fnmatch import fnmatch
from pathlib import Path

from .common import logger, HarrierProblem
from .config import Config
from .tool_chain import ToolChainFactory, ToolChain
from .tools import walk, hash_file


def build(config: Config):
    return Builder(config).build()


class Builder:
    _hash_dict = _previous_source_map = None
    _already_built = False
    _previous_hash_dict = {}
    _previous_full_build = False

    def __init__(self, config: Config):
        self._config = config
        self._tool_chain_factory = ToolChainFactory(config)
        self._exclude_patterns = self._config.exclude_patterns

    def build(self, partial=False) -> ToolChain:
        if not partial:
            self._previous_full_build = True
        elif self._previous_full_build:
            raise HarrierProblem('Partial builds are not allowed following full builds with the same builder')

        self._delete(partial)
        self._already_built = True

        tools = self._tool_chain_factory(partial)
        all_files = self._file_list()

        logger.debug('%s files to build: %s', len(all_files), ', '.join(map(str, all_files)))

        self._hash_dict = {}
        files_changed = 0

        for file_path in all_files:
            changed = self._file_changed(file_path) if partial else True
            files_changed += changed
            tools.assign_file(file_path, changed)
            if partial:
                logger.debug('%20s: %s', file_path, 'changed' if changed else 'unchanged')

        extra_files = tools.get_extra_files()
        logger.debug('%s extra files will be generated', len(extra_files))

        for file_path in extra_files:
            files_changed += 1
            tools.assign_file(file_path, True)
            logger.debug('%20s: extra file', file_path)

        if partial:
            logger.info('%s files changed or associated with changed files', files_changed)

        self._config.target_dir.mkdir(parents=True, exist_ok=True)
        tools.build()

        tool_str = 'tool' if tools.tools_run == 1 else 'tools'
        file_str = 'file' if tools.files_built == 1 else 'files'
        logger.info('Built %s %s with %s %s', tools.files_built, file_str, tools.tools_run, tool_str)

        for t in tools:
            t.cleanup()

        if partial:
            # TODO deleting stale files is fairly aggressive and could cause problems, maybe needs switch
            # could check for force activation of the tool associated with this file, but would need to
            # modify check_ownership on Jinja
            stale_files = self._delete_stale()
            logger.debug('Deleted %d stale files', stale_files)

        if partial:
            self._previous_hash_dict = copy(self._hash_dict)
            self._previous_source_map = copy(tools.source_map)
        logger.debug('-' * 20)
        return tools

    def _file_changed(self, file_path):
        file_hash = hash_file(Path(self._config.root) / file_path)

        # add hash so it can be used on next build
        self._hash_dict[file_path] = file_hash

        # check if the file_hash exists in and matches _hash_dict
        return self._previous_hash_dict.get(file_path) != file_hash

    def _file_list(self):
        all_files = walk(self._config.root)
        logger.debug('%s files in root directory', len(all_files))

        before_exclude = len(all_files)
        all_files = list(filter(self._not_excluded, all_files))
        logger.debug('%s files excluded', before_exclude - len(all_files))
        return all_files

    def _not_excluded(self, fn):
        return not any(fnmatch(str(fn), m) for m in self._exclude_patterns)

    def _delete_stale(self):
        """
        Find deleted files root by comparing hash_dict and previous hash_dict, then delete the associated target
        files.
        """
        c = 0
        for deleted_file in (set(self._previous_hash_dict.keys()) - set(self._hash_dict.keys())):
            for target_deleted_file in self._previous_source_map.get(deleted_file, []):
                self._config.target_dir.joinpath(target_deleted_file).unlink()
            c += 1
        return c

    def _delete(self, partial):
        if not self._config.target_dir.exists():
            return

        reason = None
        if not partial:
            reason = 'Full'
        elif not self._already_built:
            reason = 'First'

        if reason:
            logger.info('%s build, deleting target directory %s', reason, self._config.target_dir)
            shutil.rmtree(str(self._config.target_dir))



import time
import shlex
import subprocess
from datetime import datetime
from fnmatch import fnmatch
from multiprocessing import Process

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, FileMovedEvent

from .config import Config
from .common import logger, HarrierProblem
from .build import Builder
from .serve import serve

# specific to jetbrains I think, very annoying if not ignored
JB_BACKUP_FILE = '*___jb_???___'


class HarrierEventHandler(PatternMatchingEventHandler):
    patterns = ['*.*']
    ignore_directories = True
    wait_delay = 1

    ignore_patterns = [
        '*/.git/*',
        '*/.idea/*',
        JB_BACKUP_FILE,
    ]

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config = config
        self.build_no = 0
        self._builder = Builder(config)
        self._passing = True
        self._build_time = datetime.now()

    def on_any_event(self, event):
        if isinstance(event, FileMovedEvent):
            if fnmatch(event._src_path, JB_BACKUP_FILE) or fnmatch(event._dest_path, JB_BACKUP_FILE):
                return
        since_build = (datetime.now() - self._build_time).total_seconds()
        if since_build <= 1:
            logger.debug('%s | %0.3f seconds since last build, skipping build', event, since_build)
            return
        logger.debug('%s | %0.3f seconds since last build, building', event, since_build)
        self.build()

    def build(self):
        self._passing = None
        start = datetime.now()
        try:
            self.build_no += 1
            time.sleep(0.05)
            logger.info('change detected, rebuilding ({})...'.format(self.build_no))
            self._builder.build(partial=True)
        except Exception:
            self._passing = False
            raise
        else:
            self._passing = True
        finally:
            self._build_time = datetime.now()
            logger.info('build %d finished in %0.2fs', self.build_no, (self._build_time - start).total_seconds())

    def check_build(self):
        while not isinstance(self._passing, bool):
            time.sleep(0.1)
        if not self._passing:
            raise HarrierProblem('build failed')

    def wait(self, extra_checks):
        while True:
            time.sleep(self.wait_delay)
            self.check_build()
            if not extra_checks():
                return


class SubprocessController:
    def __init__(self, command):
        self._cmd = command
        logger.info('starting subprocess "%s"', command)
        args = shlex.split(command)
        self.p = subprocess.Popen(args)

    def terminate(self):
        if self.p.returncode is None:
            self.p.terminate()

    def check(self):
        self.p.poll()
        if self.p.returncode is None:
            return True
        if self.p.returncode == 0:
            logger.warning('subprocess "%s" exited', self._cmd)
        else:
            logger.error('subprocess "%s" exited with errors (%r)', self._cmd, self.p.returncode)


class SubprocessGroupController:
    def __init__(self, subprocess_list):
        self.subprocesses = [SubprocessController(c) for c in subprocess_list]

    def check(self):
        return all(s.check() for s in self.subprocesses)

    def terminate(self):
        [p.terminate() for p in self.subprocesses]


def watch(config: Config):
    observer = Observer()
    event_handler = HarrierEventHandler(config)
    logger.info('Watch mode starting...')
    event_handler.build()
    event_handler.check_build()

    server_process = Process(target=serve, args=(config.target_dir, config.uri_subdirectory, config.serve_port,
                                                 config.asset_file))
    server_process.start()

    sp_ctrl = SubprocessGroupController(config.subprocesses)

    observer.schedule(event_handler, str(config.root), recursive=True)
    observer.start()
    try:
        event_handler.wait(sp_ctrl.check)
    except KeyboardInterrupt:
        pass
    finally:
        logger.warning('killing dev server')
        sp_ctrl.terminate()
        observer.stop()
        observer.join()
        if server_process.exitcode not in {None, 0}:
            raise RuntimeError('Server process already terminated with exit code {}'.format(server_process.exitcode))
        else:
            server_process.terminate()
            time.sleep(0.1)



import sys
from setuptools import setup
from harrier import VERSION

description = 'Jinja2 & sass/scss aware site builder'
long_description = description

if 'sdist' in sys.argv:
    import pypandoc
    with open('README.md', 'r') as f:
        text = f.read()
    text = text[:text.find('<!-- end description -->')].strip('\n ')
    long_description = pypandoc.convert(text, 'rst', format='md')


def check_livereload_js():
    import hashlib
    from pathlib import Path
    live_reload_221_hash = 'a451e4d39b8d7ef62d380d07742b782f'
    live_reload_221_url = 'https://raw.githubusercontent.com/livereload/livereload-js/v2.2.1/dist/livereload.js'

    path = Path(__file__).absolute().parent.joinpath('harrier/livereload.js')
    if path.is_file():
        with path.open('rb') as fr:
            file_hash = hashlib.md5(fr.read()).hexdigest()
        if file_hash == live_reload_221_hash:
            return

    import urllib.request

    print('downloading livereload:\nurl:  {}\npath: {}'.format(live_reload_221_url, path))
    with urllib.request.urlopen(live_reload_221_url) as r:
        with path.open('wb') as fw:
            fw.write(r.read())

check_livereload_js()


setup(
    name='harrier',
    version=str(VERSION),
    description=description,
    long_description=long_description,
    classifiers=[
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='css,sass,scss,jinja,jinja2,build,static,static site generator',
    author='Samuel Colvin',
    author_email='s@muelcolvin.com',
    url='https://github.com/samuelcolvin/harrier',
    license='MIT',
    packages=['harrier'],
    zip_safe=True,
    package_data={'harrier': ['harrier.default.yml', 'livereload.js']},
    entry_points="""
        [console_scripts]
        harrier=harrier.cli:cli
    """,
    install_requires=[
        'Jinja2>=2.8',
        'PyYAML>=3.11',
        'click>=6.2',
        'libsass>=0.10.1',
        'watchdog>=0.8.3',
        'aiohttp>=0.21.5',
    ]
)



import logging
import asyncio
import re
import json
from pathlib import Path

import click
import aiohttp
from aiohttp.web_exceptions import HTTPNotModified, HTTPNotFound
from aiohttp import web
from aiohttp.web_urldispatcher import StaticRoute
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

logger = logging.getLogger('dev_server')

WS = 'websockets'


class DevLogHandler(logging.Handler):
    colours = {
        logging.DEBUG: 'white',
        logging.INFO: 'blue',
        logging.WARN: 'yellow',
    }

    def __init__(self, *args):
        super().__init__(*args)
        self._width = click.get_terminal_size()[0]

    def emit(self, record):
        log_entry = self.format(record)
        m = re.match('^(\[.*?\] )', log_entry)
        time = click.style(m.groups()[0], fg='magenta')
        msg = log_entry[m.end():]
        if record.levelno == logging.INFO and msg.startswith(' >'):
            msg = '{} {}'.format(click.style(' >', fg='blue'), msg[3:])
        else:
            msg = click.style(msg, fg=self.colours.get(record.levelno, 'red'))
        click.echo(time + msg)

handler = DevLogHandler()
formatter = logging.Formatter('[%(asctime)s] %(message)s', datefmt='%H:%M:%S')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)
logger.propagate = False


def serve(serve_root, subdirectory, port, asset_file=None):
    app = create_app(serve_root, subdirectory=subdirectory, asset_file=asset_file)

    # TODO in theory file watching could be replaced by accessing tool_chain.source_map
    observer = Observer()
    event_handler = DevServerEventEventHandler(app, serve_root)
    observer.schedule(event_handler, str(serve_root), recursive=True)
    observer.start()

    logger.info('Started dev server at http://localhost:%s, use Ctrl+C to quit', port)

    try:
        web.run_app(app, port=port, print=lambda msg: None)
    except KeyboardInterrupt:
        pass
    finally:
        observer.stop()
        observer.join()


def create_app(serve_root, subdirectory, asset_file, loop=None):
    loop = loop or asyncio.new_event_loop()
    app = web.Application(loop=loop)
    app[WS] = []

    app.router.add_route('GET', '/livereload.js', lr_script_handler)
    app.router.add_route('GET', '/livereload', websocket_handler)

    assert_path = asset_file and serve_root / asset_file
    serve_root = str(serve_root) + '/'
    app.router.register_route(HarrierStaticRoute('static-router', subdirectory, serve_root, assert_path=assert_path))
    if subdirectory != '/':
        app.router.add_route('*', '/{path:.*}', OutsideSubdirectory(subdirectory, assert_path))
    return app


class DevServerEventEventHandler(FileSystemEventHandler):
    def __init__(self, app, serve_root):
        super().__init__()
        self._serve_root = serve_root
        self._app = app

    def on_any_event(self, event):
        path = Path(event.src_path).relative_to(self._serve_root)
        clients = len(self._app[WS])
        logger.info('prompting reload of %s on %d client%s', path, clients, '' if clients == 1 else 's')
        for i, ws in enumerate(self._app[WS]):
            data = {
                'command': 'reload',
                'path': str(path),
                'liveCSS': True,
                'liveImg': True,
            }
            ws.send_str(json.dumps(data))


async def lr_script_handler(request):
    script_key = 'livereload_script'
    lr_script = request.app.get(script_key)
    if lr_script is None:
        lr_path = Path(__file__).absolute().parent.joinpath('livereload.js')
        with lr_path.open('rb') as f:
            lr_script = f.read()
            request.app[script_key] = lr_script
    return web.Response(body=lr_script, content_type='application/javascript')


async def websocket_handler(request):

    ws = web.WebSocketResponse()
    request.app[WS].append(ws)
    await ws.prepare(request)
    ws_type_lookup = {k.value: v for v, k in aiohttp.MsgType.__members__.items()}

    async for msg in ws:
        if msg.tp == aiohttp.MsgType.text:
            try:
                data = json.loads(msg.data)
            except json.JSONDecodeError as e:
                logger.error('JSON decode error: %s', str(e))
            else:
                command = data['command']
                if command == 'hello':
                    if 'http://livereload.com/protocols/official-7' not in data['protocols']:
                        logger.error('live reload protocol 7 not supported by client %s', msg.data)
                        ws.close()
                    else:
                        handshake = {
                            'command': 'hello',
                            'protocols': [
                                'http://livereload.com/protocols/official-7',
                            ],
                            'serverName': 'livereload-aiohttp',
                        }
                        ws.send_str(json.dumps(handshake))
                elif command == 'info':
                    logger.info('browser connected at %s', data['url'])
                    logger.debug('browser plugins: %s', data['plugins'])
                else:
                    logger.error('Unknown ws message %s', msg.data)
        elif msg.tp == aiohttp.MsgType.error:
            logger.error('ws connection closed with exception %s',  ws.exception())
        else:
            logger.error('unknown websocket message type %s, data: %s', ws_type_lookup[msg.tp], msg.data)

    # TODO gracefully close websocket connections on app shutdown
    logger.debug('browser disconnected')
    request.app[WS].remove(ws)

    return ws


class HarrierStaticRoute(StaticRoute):
    def __init__(self, *args, **kwargs):
        self._asset_path = kwargs.pop('assert_path', None)
        super().__init__(*args, **kwargs)

    async def handle(self, request):
        filename = request.match_info['filename']
        try:
            filepath = self._directory.joinpath(filename).resolve()
        except (ValueError, FileNotFoundError, OSError):
            pass
        else:
            if filepath.is_dir():
                request.match_info['filename'] = str(filepath.joinpath('index.html').relative_to(self._directory))
        status, length = 'unknown', ''
        try:
            response = await super().handle(request)
        except HTTPNotModified:
            status, length = 304, 0
            raise
        except HTTPNotFound:
            _404_msg = '404: Not Found\n\n' + _get_asset_content(self._asset_path)
            response = web.Response(body=_404_msg.encode('utf8'), status=404)
            status, length = response.status, response.content_length
        else:
            status, length = response.status, response.content_length
        finally:
            l = logger.info if status in {200, 304} else logger.warning
            l(' > %s %s %s %s', request.method, request.path, status, _fmt_size(length))
        return response


class OutsideSubdirectory:
    def __init__(self, prefix, assert_path):
        self._asset_path = assert_path
        self._msg = '404: Not Found (files are being served from the subdirectory "{}" only)\n\n'.format(prefix)

    async def __call__(self, request):
        msg = self._msg + _get_asset_content(self._asset_path)
        r = web.Response(body=msg.encode('utf8'), status=404)
        logger.warning(' > %s %s %s %s', request.method, request.path, r.status, _fmt_size(r.content_length))
        return r


def _get_asset_content(asset_path):
    if not asset_path:
        return ''
    with asset_path.open() as f:
        return 'Asset file contents:\n\n{}'.format(f.read())


def _fmt_size(num):
    if num == '':
        return ''
    if num < 1024:
        return '{:0.0f}B'.format(num)
    else:
        return "{:0.1f}KB".format(num / 1024)



import sys
from itertools import chain
from operator import attrgetter
from importlib import import_module

from .common import logger


class _ToolList(list):
    def prioritise(self, attr):
        super().sort(key=attrgetter(attr), reverse=True)


class ToolChain(_ToolList):
    files_built, tools_run, source_map = 0, 0, None

    def __init__(self, tool_classes, config, partial):
        super().__init__(t(config, partial) for t in tool_classes)

    def get_extra_files(self) -> set:
        return set(chain(*[t.extra_files for t in self if t.active]))

    def sort_on(self, attr, reverse=False):
        super().sort(key=attrgetter(attr), reverse=reverse)

    def assign_file(self, file_path, changed):
        for t in self:
            if t.assign_file(file_path, changed):
                break

    def build(self):
        self.prioritise('build_priority')
        self.source_map = {}
        for t in self:
            if not t.active:
                continue
            files_built, source_map = t.build()
            logger.debug('built %s files with %s', files_built, t.name)
            self.files_built += files_built
            self.source_map.update(source_map)
            self.tools_run += 1

    @property
    def status(self):
        return {
            'tools': len(self),
            'tools_run': self.tools_run,
            'files_built': self.files_built
        }

    def get_tool(self, tool_name):
        return next((t for t in self if t.name.lower() == tool_name.lower()), None)

    def __str__(self):
        return '{tools} tools of which {tools_run} run, {files_built} files built'.format(**self.status)


class ToolChainFactory(_ToolList):
    def __init__(self, config):
        self._config = config
        super().__init__(import_string(t) for t in config.tools)
        self.prioritise('ownership_priority')

    def __call__(self, partial) -> ToolChain:
        return ToolChain(self, self._config, partial)


def import_string(dotted_path):
    """
    Stolen verbatim from django.

    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:  # pragma: no cover
        e = ImportError("{} doesn't look like a module path".format(dotted_path))
        raise e.with_traceback(sys.exc_info()[2])

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:  # pragma: no cover
        e = ImportError('Module "{}" does not define a "{}" attribute/class'.format(module_path, class_name))
        raise e.with_traceback(sys.exc_info()[2])
